from fast_tts.fast_tts import FastTTS





FastTTS.fast_tts('Hello World! This is the ultimate test of the Fast TTS! I hope it works! Thank you!', 'C:\\backend_server\\ServerFiles\\Speakers\\Drinker', 'C:\\Users\\mathi\\OneDrive\\Bureau')